@@ -1,6 +1,6 @@
 // =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
 // Mobius Forensic Toolkit
-// Copyright (C) 2008,2009,2010,2011,2012,2013,2014,2015,2016,2017,2018 Eduardo Aguiar
+// Copyright (C) 2008,2009,2010,2011,2012,2013,2014,2015,2016,2017,2018,2019 Eduardo Aguiar
 //
 // This program is free software; you can redistribute it and/or modify it
 // under the terms of the GNU General Public License as published by the
@@ -20,11 +20,37 @@
 //! \brief C++ API module wrapper
 //! \author Eduardo Aguiar
 // =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
+#include <pymobius.h>
 #include "entry.h"
 #include "stream.h"
-#include "api_datetime.h"
 #include "io/reader.h"
 
+// =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
+//! \brief Create <i>entry</i> Python object from C++ object
+//! \param obj C++ object
+//! \return new Python object
+// =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
+PyObject *
+PyMobius_Filesystem_Entry_from_cpp (mobius::filesystem::entry obj)
+{
+  PyObject *ret = nullptr;
+
+  if (obj)
+    {
+      ret = _PyObject_New (&filesystem_entry_t);
+
+      if (ret)
+        ((filesystem_entry_o *) ret)->obj = new mobius::filesystem::entry (obj);
+    }
+  else
+    {
+      Py_INCREF (Py_None);
+      ret = Py_None;
+    }
+
+  return ret;
+}
+      
 // =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
 //! \brief tp_new (default constructor)
 //! \param type type object
@@ -59,7 +85,18 @@ tp_dealloc (filesystem_entry_o *self)
 static PyObject *
 tp_getter_size (filesystem_entry_o *self)
 {
-  return PyLong_FromUnsignedLongLong (self->obj->get_size ());
+  PyObject *ret = nullptr;
+
+  try
+    {
+      ret = PyLong_FromUnsignedLongLong (self->obj->get_size ());
+    }
+  catch (const std::exception& e)
+    {
+      PyErr_SetString (PyExc_Exception, e.what ());
+    }
+
+  return ret;
 }
 
 // =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
@@ -68,7 +105,18 @@ tp_getter_size (filesystem_entry_o *self)
 static PyObject *
 tp_getter_type (filesystem_entry_o *self)
 {
-  return PyInt_FromLong (int (self->obj->get_type ()));
+  PyObject *ret = nullptr;
+
+  try
+    {
+      ret = PyLong_FromLong (int (self->obj->get_type ()));
+    }
+  catch (const std::exception& e)
+    {
+      PyErr_SetString (PyExc_Exception, e.what ());
+    }
+
+  return ret;
 }
 
 // =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
@@ -77,7 +125,18 @@ tp_getter_type (filesystem_entry_o *self)
 static PyObject *
 tp_getter_inode (filesystem_entry_o *self)
 {
-  return PyLong_FromUnsignedLongLong (self->obj->get_inode ());
+  PyObject *ret = nullptr;
+
+  try
+    {
+      ret = PyLong_FromUnsignedLongLong (self->obj->get_inode ());
+    }
+  catch (const std::exception& e)
+    {
+      PyErr_SetString (PyExc_Exception, e.what ());
+    }
+
+  return ret;
 }
 
 // =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
@@ -86,7 +145,18 @@ tp_getter_inode (filesystem_entry_o *self)
 static PyObject *
 tp_getter_mode (filesystem_entry_o *self)
 {
-  return PyInt_FromLong (self->obj->get_mode ());
+  PyObject *ret = nullptr;
+
+  try
+    {
+      ret = PyLong_FromLong (self->obj->get_mode ());
+    }
+  catch (const std::exception& e)
+    {
+      PyErr_SetString (PyExc_Exception, e.what ());
+    }
+
+  return ret;
 }
 
 // =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
@@ -95,7 +165,18 @@ tp_getter_mode (filesystem_entry_o *self)
 static PyObject *
 tp_getter_uid (filesystem_entry_o *self)
 {
-  return PyInt_FromLong (self->obj->get_uid ());
+  PyObject *ret = nullptr;
+
+  try
+    {
+      ret = PyLong_FromLong (self->obj->get_uid ());
+    }
+  catch (const std::exception& e)
+    {
+      PyErr_SetString (PyExc_Exception, e.what ());
+    }
+
+  return ret;
 }
 
 // =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
@@ -104,7 +185,18 @@ tp_getter_uid (filesystem_entry_o *self)
 static PyObject *
 tp_getter_gid (filesystem_entry_o *self)
 {
-  return PyInt_FromLong (self->obj->get_gid ());
+  PyObject *ret = nullptr;
+
+  try
+    {
+      ret = PyLong_FromLong (self->obj->get_gid ());
+    }
+  catch (const std::exception& e)
+    {
+      PyErr_SetString (PyExc_Exception, e.what ());
+    }
+
+  return ret;
 }
 
 // =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
@@ -113,7 +205,18 @@ tp_getter_gid (filesystem_entry_o *self)
 static PyObject *
 tp_getter_path (filesystem_entry_o *self)
 {
-  return PyString_FromString (self->obj->get_path ().c_str ());
+  PyObject *ret = nullptr;
+
+  try
+    {
+      ret = PyString_from_std_string (self->obj->get_path ());
+    }
+  catch (const std::exception& e)
+    {
+      PyErr_SetString (PyExc_Exception, e.what ());
+    }
+
+  return ret;
 }
 
 // =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
@@ -135,8 +238,15 @@ tp_setter_path (filesystem_entry_o *self, PyObject *value, void *)
     }
 
   // set path
-  const char *arg_path = PyString_AsString (value);
-  self->obj->set_path (arg_path);
+  try
+    {
+      self->obj->set_path (PyString_as_std_string (value));
+    }
+  catch (const std::exception& e)
+    {
+      PyErr_SetString (PyExc_Exception, e.what ());
+      return -1;
+    }
 
   return 0;
 }
@@ -147,7 +257,18 @@ tp_setter_path (filesystem_entry_o *self, PyObject *value, void *)
 static PyObject *
 tp_getter_name (filesystem_entry_o *self)
 {
-  return PyString_FromString (self->obj->get_name ().c_str ());
+  PyObject *ret = nullptr;
+
+  try
+    {
+      ret = PyString_from_std_string (self->obj->get_name ());
+    }
+  catch (const std::exception& e)
+    {
+      PyErr_SetString (PyExc_Exception, e.what ());
+    }
+
+  return ret;
 }
 
 // =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
@@ -156,7 +277,18 @@ tp_getter_name (filesystem_entry_o *self)
 static PyObject *
 tp_getter_short_name (filesystem_entry_o *self)
 {
-  return PyString_FromString (self->obj->get_short_name ().c_str ());
+  PyObject *ret = nullptr;
+
+  try
+    {
+      ret = PyString_from_std_string (self->obj->get_short_name ());
+    }
+  catch (const std::exception& e)
+    {
+      PyErr_SetString (PyExc_Exception, e.what ());
+    }
+
+  return ret;
 }
 
 // =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
@@ -165,7 +297,18 @@ tp_getter_short_name (filesystem_entry_o *self)
 static PyObject *
 tp_getter_creation_time (filesystem_entry_o *self)
 {
-  return PyDateTime_from_datetime (self->obj->get_creation_time ());
+  PyObject *ret = nullptr;
+
+  try
+    {
+      ret = PyDateTime_from_datetime (self->obj->get_creation_time ());
+    }
+  catch (const std::exception& e)
+    {
+      PyErr_SetString (PyExc_Exception, e.what ());
+    }
+
+  return ret;
 }
 
 // =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
@@ -174,7 +317,18 @@ tp_getter_creation_time (filesystem_entry_o *self)
 static PyObject *
 tp_getter_last_modification_time (filesystem_entry_o *self)
 {
-  return PyDateTime_from_datetime (self->obj->get_last_modification_time ());
+  PyObject *ret = nullptr;
+
+  try
+    {
+      ret = PyDateTime_from_datetime (self->obj->get_last_modification_time ());
+    }
+  catch (const std::exception& e)
+    {
+      PyErr_SetString (PyExc_Exception, e.what ());
+    }
+
+  return ret;
 }
 
 // =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
@@ -183,7 +337,18 @@ tp_getter_last_modification_time (filesystem_entry_o *self)
 static PyObject *
 tp_getter_last_access_time (filesystem_entry_o *self)
 {
-  return PyDateTime_from_datetime (self->obj->get_last_access_time ());
+  PyObject *ret = nullptr;
+
+  try
+    {
+      ret = PyDateTime_from_datetime (self->obj->get_last_access_time ());
+    }
+  catch (const std::exception& e)
+    {
+      PyErr_SetString (PyExc_Exception, e.what ());
+    }
+
+  return ret;
 }
 
 // =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
@@ -192,7 +357,18 @@ tp_getter_last_access_time (filesystem_entry_o *self)
 static PyObject *
 tp_getter_last_metadata_time (filesystem_entry_o *self)
 {
-  return PyDateTime_from_datetime (self->obj->get_last_metadata_time ());
+  PyObject *ret = nullptr;
+
+  try
+    {
+      ret = PyDateTime_from_datetime (self->obj->get_last_metadata_time ());
+    }
+  catch (const std::exception& e)
+    {
+      PyErr_SetString (PyExc_Exception, e.what ());
+    }
+
+  return ret;
 }
 
 // =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
@@ -201,7 +377,18 @@ tp_getter_last_metadata_time (filesystem_entry_o *self)
 static PyObject *
 tp_getter_deletion_time (filesystem_entry_o *self)
 {
-  return PyDateTime_from_datetime (self->obj->get_deletion_time ());
+  PyObject *ret = nullptr;
+
+  try
+    {
+      ret = PyDateTime_from_datetime (self->obj->get_deletion_time ());
+    }
+  catch (const std::exception& e)
+    {
+      PyErr_SetString (PyExc_Exception, e.what ());
+    }
+
+  return ret;
 }
 
 // =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
@@ -210,7 +397,18 @@ tp_getter_deletion_time (filesystem_entry_o *self)
 static PyObject *
 tp_getter_is_deleted (filesystem_entry_o *self)
 {
-  return PyBool_FromLong (self->obj->is_deleted ());
+  PyObject *ret = nullptr;
+
+  try
+    {
+      ret = PyBool_FromLong (self->obj->is_deleted ());
+    }
+  catch (const std::exception& e)
+    {
+      PyErr_SetString (PyExc_Exception, e.what ());
+    }
+
+  return ret;
 }
 
 // =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
@@ -219,7 +417,18 @@ tp_getter_is_deleted (filesystem_entry_o *self)
 static PyObject *
 tp_getter_is_reallocated (filesystem_entry_o *self)
 {
-  return PyBool_FromLong (self->obj->is_reallocated ());
+  PyObject *ret = nullptr;
+
+  try
+    {
+      ret = PyBool_FromLong (self->obj->is_reallocated ());
+    }
+  catch (const std::exception& e)
+    {
+      PyErr_SetString (PyExc_Exception, e.what ());
+    }
+
+  return ret;
 }
 
 // =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
@@ -332,34 +541,23 @@ static PyGetSetDef tp_getset[] =
 static PyObject *
 tp_f_get_children (filesystem_entry_o *self, PyObject *)
 {
-  PyObject *ret = PyList_New (0);
-  if (!ret)
-    return nullptr;
+  PyObject *ret = nullptr;
+
+  // execute C++ code
+  PyThreadState *_save = PyEval_SaveThread ();
 
-  // populate list
   try
     {
-      for (auto child : self->obj->get_children ())
-        {
-          filesystem_entry_o *item = (filesystem_entry_o *) _PyObject_New (&filesystem_entry_t);
-
-          if (!item)
-            {
-              Py_CLEAR (ret);
-              return nullptr;
-            }
-
-          item->obj = new mobius::filesystem::entry (child);
-          PyList_Append (ret, (PyObject *) item);
-        }
+      auto children = self->obj->get_children ();
+      PyEval_RestoreThread (_save);
+      ret = PyList_from_cpp_container (children, PyMobius_Filesystem_Entry_from_cpp);
     }
   catch (const std::exception& e)
     {
+      PyEval_RestoreThread (_save);
       PyErr_SetString (PyExc_Exception, e.what ());
-      Py_CLEAR (ret);
-      ret = nullptr;
     }
-
+  
   return ret;
 }
 
@@ -369,26 +567,9 @@ tp_f_get_children (filesystem_entry_o *self, PyObject *)
 static PyObject *
 tp_f_get_parent (filesystem_entry_o *self, PyObject *)
 {
-  // execute C++ code
-  PyObject *ret = nullptr;
   auto parent = self->obj->get_parent ();
 
-  // if parent is found, create object, otherwise returns None
-  if (parent)
-    {
-      ret = _PyObject_New (&filesystem_entry_t);
-
-      if (ret)
-        ((filesystem_entry_o *) ret)->obj = new mobius::filesystem::entry (parent);
-    }
-
-  else
-    {
-      Py_INCREF (Py_None);
-      ret = Py_None;
-    }
-
-  return ret;
+  return PyMobius_Filesystem_Entry_from_cpp (parent);
 }
 
 // =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
@@ -411,26 +592,11 @@ tp_f_get_child_by_name (filesystem_entry_o *self, PyObject *args)
     {
       auto child = self->obj->get_child_by_name (arg_name, bool (arg_case_sensitive));
 
-      // if child is found, create object, otherwise returns None
-      if (child)
-        {
-          ret = _PyObject_New (&filesystem_entry_t);
-
-          if (ret)
-            ((filesystem_entry_o *) ret)->obj = new mobius::filesystem::entry (child);
-        }
-
-      else
-        {
-          Py_INCREF (Py_None);
-          ret = Py_None;
-        }
+      ret = PyMobius_Filesystem_Entry_from_cpp (child);
     }
   catch (const std::exception& e)
     {
       PyErr_SetString (PyExc_Exception, e.what ());
-      Py_CLEAR (ret);
-      ret = nullptr;
     }
 
   return ret;
@@ -456,26 +622,11 @@ tp_f_get_child_by_path (filesystem_entry_o *self, PyObject *args)
     {
       auto child = self->obj->get_child_by_path (arg_path, bool (arg_case_sensitive));
 
-      // if child is found, create object, otherwise returns None
-      if (child)
-        {
-          ret = _PyObject_New (&filesystem_entry_t);
-
-          if (ret)
-            ((filesystem_entry_o *) ret)->obj = new mobius::filesystem::entry (child);
-        }
-
-      else
-        {
-          Py_INCREF (Py_None);
-          ret = Py_None;
-        }
+      ret = PyMobius_Filesystem_Entry_from_cpp (child);
     }
   catch (const std::exception& e)
     {
       PyErr_SetString (PyExc_Exception, e.what ());
-      Py_CLEAR (ret);
-      ret = nullptr;
     }
 
   return ret;
@@ -487,31 +638,18 @@ tp_f_get_child_by_path (filesystem_entry_o *self, PyObject *args)
 static PyObject *
 tp_f_get_streams (filesystem_entry_o *self, PyObject *)
 {
-  PyObject *ret = PyList_New (0);
-  if (!ret)
-    return nullptr;
+  PyObject *ret = nullptr;
 
   try
     {
-      for (auto stream : self->obj->get_streams ())
-        {
-          filesystem_stream_o *item = (filesystem_stream_o *) _PyObject_New (&filesystem_stream_t);
-
-          if (!item)
-            {
-              Py_CLEAR (ret);
-              return nullptr;
-            }
-
-          item->obj = new mobius::filesystem::stream (stream);
-          PyList_Append (ret, (PyObject *) item);
-        }
+      ret = PyList_from_cpp_container (
+               self->obj->get_streams (),
+               PyMobius_Filesystem_Stream_from_cpp
+            );
     }
   catch (const std::exception& e)
     {
       PyErr_SetString (PyExc_Exception, e.what ());
-      Py_CLEAR (ret);
-      ret = nullptr;
     }
 
   return ret;
@@ -523,7 +661,18 @@ tp_f_get_streams (filesystem_entry_o *self, PyObject *)
 static PyObject *
 tp_f_new_reader (filesystem_entry_o *self, PyObject *)
 {
-  return PyMobius_IO_Reader_from_cpp (self->obj->new_reader ());
+  PyObject *ret = nullptr;
+
+  try
+    {
+      ret = PyMobius_IO_Reader_from_cpp (self->obj->new_reader ());
+    }
+  catch (const std::exception& e)
+    {
+      PyErr_SetString (PyExc_Exception, e.what ());
+    }
+
+  return ret;
 }
 
 // =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
@@ -532,7 +681,18 @@ tp_f_new_reader (filesystem_entry_o *self, PyObject *)
 static PyObject *
 tp_f_is_folder (filesystem_entry_o *self, PyObject *)
 {
-  return PyBool_FromLong (self->obj->is_folder ());
+  PyObject *ret = nullptr;
+
+  try
+    {
+      ret = PyBool_FromLong (self->obj->is_folder ());
+    }
+  catch (const std::exception& e)
+    {
+      PyErr_SetString (PyExc_Exception, e.what ());
+    }
+
+  return ret;
 }
 
 // =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
@@ -541,7 +701,18 @@ tp_f_is_folder (filesystem_entry_o *self, PyObject *)
 static PyObject *
 tp_f_is_file (filesystem_entry_o *self, PyObject *)
 {
-  return PyBool_FromLong (self->obj->is_file ());
+  PyObject *ret = nullptr;
+
+  try
+    {
+      ret = PyBool_FromLong (self->obj->is_file ());
+    }
+  catch (const std::exception& e)
+    {
+      PyErr_SetString (PyExc_Exception, e.what ());
+    }
+
+  return ret;
 }
 
 // =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
